# bring definitions to the top level
from ds_capability.components.feature_select import FeatureSelect
from ds_capability.components.feature_engineer import FeatureEngineer
from ds_capability.components.feature_transform import FeatureTransform
from ds_capability.components.feature_build import FeatureBuild
from ds_capability.components.automl import AutoML
from ds_capability.components.controller import Controller


# release version number picked up in the setup.py
__version__ = '0.20.5'
